searchNodes=[{"ref":"Commanded.Projections.Ecto.html","title":"Commanded.Projections.Ecto","type":"module","doc":"Read model projections for Commanded using Ecto. Example usage defmodule Projector do use Commanded.Projections.Ecto, application: MyApp.Application, name: &quot;my-projection&quot;, repo: MyRepo, schema_prefix: &quot;my-prefix&quot;, timeout: :infinity project %Event{}, _metadata, fn multi -&gt; Ecto.Multi.insert(multi, :my_projection, %MyProjection{...}) end project %AnotherEvent{}, fn multi -&gt; Ecto.Multi.insert(multi, :my_projection, %MyProjection{...}) end end Guides Getting started Usage"},{"ref":"Commanded.Projections.Ecto.html#project/2","title":"Commanded.Projections.Ecto.project/2","type":"macro","doc":""},{"ref":"Commanded.Projections.Ecto.html#project/3","title":"Commanded.Projections.Ecto.project/3","type":"macro","doc":""},{"ref":"changelog.html","title":"Changelog","type":"extras","doc":"Changelog"},{"ref":"changelog.html#v1-0-0-rc-0","title":"Changelog - v1.0.0-rc.0","type":"extras","doc":"Add .formatter.exs to Hex package (#19). Add microseconds to timestamp fields in projection_versions (#22). Support multiple Commanded apps (#25)."},{"ref":"changelog.html#v0-8-0","title":"Changelog - v0.8.0","type":"extras","doc":"Enhancements Upgrade to Ecto v3 (#17). Use lambda instead of unhygienic var in projection macros (#13). Previously magic multi: project %AnEvent{name: name}, _metadata do Ecto.Multi.insert(multi, :example_projection, %ExampleProjection{name: name}) end Now multi is provided as a argument to the project function: project %AnEvent{name: name}, _metadata, fn multi -&gt; Ecto.Multi.insert(multi, :example_projection, %ExampleProjection{name: name}) end The previous do block approach is still supported, but has been deprecated. It will be removed in the next release."},{"ref":"changelog.html#v0-7-1","title":"Changelog - v0.7.1","type":"extras","doc":"Bug fixes Ensure errors encountered while building the Ecto.Multi data structure within a project function are caught and passed to the error/3 callback."},{"ref":"changelog.html#v0-7-0","title":"Changelog - v0.7.0","type":"extras","doc":"Enhancements Support Commanded&#39;s event handler error/3 callback (#12)."},{"ref":"changelog.html#v0-6-0","title":"Changelog - v0.6.0","type":"extras","doc":"Enhancements Pass through any additional projector configuration options to Commanded event handler. Allows new Commanded features to be used without updating this library (e.g. specify consistency option)."},{"ref":"changelog.html#v0-5-0","title":"Changelog - v0.5.0","type":"extras","doc":"Enhancements Allow an Ecto schema prefix to be defined in config or per handler (#4)."},{"ref":"changelog.html#v0-4-0","title":"Changelog - v0.4.0","type":"extras","doc":"Enhancements Add repo option to Commanded.Projections.Ecto macro (#1). Optional after_update/3 callback function in projectors."},{"ref":"getting-started.html","title":"Getting started","type":"extras","doc":"Getting started You should already have Ecto installed and configured before proceeding. Please follow Ecto&#39;s Getting Started guide to get going first. Add commanded_ecto_projections to your list of dependencies in mix.exs: def deps do [{:commanded_ecto_projections, &quot;~&gt; 1.0.0-rc.0&quot;}] end Generate an Ecto migration in your app: $ mix ecto.gen.migration create_projection_versions Modify the generated migration, in priv/repo/migrations, to create the projection_versions table: defmodule CreateProjectionVersions do use Ecto.Migration def change do create table(:projection_versions, primary_key: false) do add(:projection_name, :text, primary_key: true) add(:last_seen_event_number, :bigint) timestamps(type: :naive_datetime_usec) end end end Run the Ecto migration: $ mix ecto.migrate Define your first read model projector: defmodule MyApp.ExampleProjector do use Commanded.Projections.Ecto, application: MyApp.Application, repo: MyApp.Projections.Repo, name: &quot;example_projection&quot; end Refer to the Usage guide for more detail on how to configure and use a read model projector."},{"ref":"usage.html","title":"Usage","type":"extras","doc":"Usage"},{"ref":"usage.html#creating-a-read-model","title":"Usage - Creating a read model","type":"extras","doc":"Use Ecto schemas to define your read model: defmodule ExampleProjection do use Ecto.Schema schema &quot;example_projections&quot; do field(:name, :string) end end"},{"ref":"usage.html#creating-a-projector","title":"Usage - Creating a projector","type":"extras","doc":"For each read model you will need to define a module that uses the Commanded.Projections.Ecto module and projects the appropriate domain events with the project macro. You must specify the following options when defining an Ecto projector: :application - (module) the Commanded application (e.g. MyApp.Application). :repo - (module) an Ecto repo (e.g. MyApp.Projections.Repo). :name - (string) a unique name used to identify the event store subscription used by the projector. Once a projector has been deployed you should not change its name. Doing so will cause a new event store subscription to be created and replay all existing events. Example defmodule MyApp.ExampleProjector do use Commanded.Projections.Ecto, application: MyApp.Application, repo: MyApp.Projections.Repo, name: &quot;example_projection&quot; project %AnEvent{name: name}, _metadata, fn multi -&gt; Ecto.Multi.insert(multi, :example_projection, %ExampleProjection{name: name}) end project %AnotherEvent{name: name}, fn multi -&gt; Ecto.Multi.insert(multi, :example_projection, %ExampleProjection{name: name}) end end Note: A read model projector is just a specialised Commanded event handler. Using the project macro The project/3 macro expects the domain event, metadata, and a single-arity function that takes and returns an Ecto.Multi data structure for grouping multiple Repo operations. These will all be executed within a single transaction. You can use Ecto.Multi to insert, update, and delete data. Examples Project an event and its metadata into a read model with project/3: project %AnEvent{name: name}, metadata, fn multi -&gt; projection = %ExampleProjection{name: name, metadata: metadata} Ecto.Multi.insert(multi, :example_projection, projection) end Use project/2 if you do not need to use the event metadata: project %AnotherEvent{name: name}, fn multi -&gt; Ecto.Multi.insert(multi, :example_projection, %ExampleProjection{name: name}) end If you want to skip a projection event, you can return the multi transaction without further modifying it: project %ItemUpdated{uuid: uuid} = event, _metadata, fn multi -&gt; case Repo.get(ItemProjection, uuid) do nil -&gt; multi item -&gt; Ecto.Multi.update(multi, :item, update_changeset(event, item)) end end"},{"ref":"usage.html#supervision","title":"Usage - Supervision","type":"extras","doc":"Your projector module must be included in your application supervision tree: defmodule MyApp.Projections.Supervisor do use Supervisor def start_link(init_arg) do Supervisor.start_link(__MODULE__, init_arg, name: __MODULE__) end @impl true def init(_init_arg) do children = [ MyApp.ExampleProjector ] Supervisor.init(children, strategy: :one_for_one) end end Warning: You should implement an error handling strategy in your projector module when supervising to prevent problematic events from causing cascading errors due too many restarts."},{"ref":"usage.html#error-handling","title":"Usage - Error handling","type":"extras","doc":"error/3 callback The Commanded.Projections.Ecto macro defines a Commanded event handler which means you can take advantage of the error/3 callback function to handle any errors returned from a project function. The error function is passed the error returned by the event handler (e.g. {:error, error}), the event causing the error, and a context map containing state passed between retries. Use the context map to track any transient state you need to access between retried failures, such as the number of failed attempts. You can return one of the following responses depending upon the error severity: {:retry, context} - retry the failed event, provide a context map containing any state passed to subsequent failures. This could be used to count the number of failures, stopping after too many. {:retry, delay, context} - retry the failed event, after sleeping for the requested delay (in milliseconds). Context is a map as described in {:retry, context} above. :skip - skip the failed event by acknowledging receipt. {:stop, reason} - stop the projector with the given reason. Error handling example Here&#39;s an example projector module where an error tagged tuple is explicitly returned from a project function, but you can also handle exceptions caused by faulty Ecto.Multi database operations in a similar manner since the errors are caught and returned as tagged tuples (e.g. {:error, %Ecto.ConstraintError{}}). defmodule MyApp.ExampleProjector do use Commanded.Projections.Ecto, application: MyApp.Application, repo: MyApp.Projections.Repo,, name: &quot;MyApp.ExampleProjector&quot; require Logger alias Commanded.Event.FailureContext project %AnEvent{}, fn _multi -&gt; {:error, :failed} end def error({:error, :failed}, %AnEvent{}, %FailureContext{}) do :skip end def error({:error, %Ecto.ConstraintError{} = error}, _event, _failure_context) do Logger.error(fn -&gt; &quot;Failed due to constraint error: &quot; &lt;&gt; inspect(error) end) :skip end def error({:error, _error}, _event, _failure_context) do :skip end end after_update/3 callback You can define an after_update/3 callback function in a projector to be called after each projected event. The function receives the event, its metadata, and all changes from the Ecto.Multi struct that were executed within the database transaction. defmodule MyApp.ExampleProjector do use Commanded.Projections.Ecto, application: MyApp.Application, repo: MyApp.Projections.Repo, name: &quot;MyApp.ExampleProjector&quot; project %AnEvent{name: name}, fn multi -&gt; Ecto.Multi.insert(multi, :example_projection, %ExampleProjection{name: name}) end def after_update(event, metadata, changes) do # Use the event, metadata, or [`Ecto.Multi`](https://hexdocs.pm/ecto/3.2.3/Ecto.Multi.html) changes and return `:ok` :ok end end You could use this function to notify subscribers that the read model has been updated (e.g. pub/sub to Phoenix channels)."},{"ref":"usage.html#schema-prefix","title":"Usage - Schema prefix","type":"extras","doc":"When using a prefix for your Ecto schemas you might also want to change the prefix for the ProjectionVersion schema. There are two options to do this: Provide a global prefix via the config: config :commanded_ecto_projections, schema_prefix: &quot;example_schema_prefix&quot; Provide the prefix to an individual projection: defmodule MyApp.ExampleProjector do use Commanded.Projections.Ecto, application: MyApp.Application, repo: MyApp.Projections.Repo, name: &quot;example_projection&quot;, schema_prefix: &quot;example_schema_prefix&quot; end"},{"ref":"usage.html#rebuilding-a-projection","title":"Usage - Rebuilding a projection","type":"extras","doc":"The projection_versions table is used to ensure that events are only projected once. To rebuild a projection you will need to: Delete the row containing the last seen event for the projection name: delete from projection_versions where projection_name = &#39;example_projection&#39;; Truncate the tables that are being populated by the projection, and restart their identity: truncate table example_projections, other_projections restart identity; You will also need to reset the event store subscription for the commanded event handler. This is specific to whichever event store you are using."}]